c===========================================================
c robinbc                                       Scott Hawley
c
c  Apply Robin Boundary condition around outer boundary
c
c===========================================================
      subroutine robinbc(u,nx,ny,nz,x,y,z,mask)

      implicit none

      include 'mgprob.inc'

      integer       nx,            ny,            nz
      real*8        u(nx,ny,nz)
      real*8        x(nx),         y(ny),         z(nz)
      integer       mask(nx,ny,nz)

c   Local variables
      real*8         h
      integer        i,              j,             k

      if (excttype .eq. 1) then
          call clcu2(u,x,z,y,nx,ny,nz,
     &           mask,.false.,.false.,.false.,.false.,.true.)

         return
      endif

      if ((nx .ne. ny) .or. (nx .ne. nz) .or. (ny .ne. nz)) then
         write(0,*) 'Error: Domain must be a cube.'
         return
      endif

      h = x(2) - x(1)
      if ((y(2)-y(1) .ne. h) .or. (z(2)-z(1).ne.h) ) then
         write(0,*) 'Error: Must have the same spacing in x, y & z'
         write(0,*) ' h = ',h
         write(0,*) ' x(2) , x(1) = ',x(2) , x(1) 
         write(0,*) ' y(2) - y(1) = ',y(2) - y(1) 
         write(0,*) ' z(2) - z(1) = ',z(2) - z(1) 
         stop
         return
      endif      

      call rbc_copy(u,nx,ny,nz,x,y,z,h)
c      call rbc_2ocopy(u,nx,ny,nz,x,y,z,h)
c      call rbc_project(u,nx,ny,nz,x,y,z,h)
c      call rbc_dirichlet(u,nx,ny,nz,x,y,z,h,mask)
c      call rbc_diffsonn(u,nx,ny,nz,x,y,z,h)
      
      return
      end 
     

c===========================================================
c   rbc_copy                                   Scott Hawley
c
c   The Robin condition is   
c
c                     partial u / partial r = (1-u)/r
c
c                           or
c
c                   [ r (u-1) ],r = 0
c
c   We enforce this by copying "in the r direction" from
c   the furthest interior points to the boundaries such that
c   the quantity r(u-1) doesn't change in the r direction
c   thus
c             r_b (u_b - 1) = r_i (u_i - 1)
c 
c                           or
c
c             u_b = 1 + (u_i - 1) * r_i / r_b
c===========================================================
      subroutine rbc_copy(u,nx,ny,nz,x,y,z,h
     &                  )
     
      implicit none

      integer       nx,            ny,            nz
      real*8        u(nx,ny,nz)
      real*8        x(nx),        y(ny),       z(nz)
      real*8         h

      integer        i,             j,            k
      real*8         r_i,           r_b

c=============================
c FACES (6 of them)
c=============================
c------------------------------------------------------------
c (interior of) +x face
c------------------------------------------------------------
      i = nx
      do k = 2, nz-1  
        do j = 2, ny-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)-h)**2 + y(j)**2 + z(k)**2)
           u(i,j,k) = 1.0d0 + (u(i-1,j,k) - 1.0d0) * r_i / r_b
c           if ((k.eq.nz/2+1).and.(j.eq.ny/2+1)) then
c              write(0,*) 'robin3dbc: u(i-1,j,k) = ',u(i-1,j,k)
c              write(0,*) 'robin3dbc: u(i,j,k) = ',u(i,j,k)
c              write(0,*) 'robin3dbc: x(i) = ',x(i)
c              write(0,*) 'robin3dbc: r_i, r_b = ',r_i, r_b
c           endif
        enddo
      enddo
      
     
c------------------------------------------------------------
c -x face
c------------------------------------------------------------
      i = 1
      do k = 2, nz-1
        do j = 2, ny-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)+h)**2 + y(j)**2 + z(k)**2)
           u(i,j,k) = 1.0d0 + (u(i+1,j,k) - 1.0d0) * r_i / r_b
        enddo
      enddo
c-----------------------------------------------------------
c +y face
c------------------------------------------------------------
      j = ny
      do k = 2, nz-1
        do i = 2, nx-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt(x(i)**2 + (y(j)-h)**2 + z(k)**2)
           u(i,j,k) = 1.0d0 + (u(i,j-1,k) - 1.0d0) * r_i / r_b
        enddo
      enddo

c------------------------------------------------------------
c -y face
c------------------------------------------------------------
      j = 1
      do k = 2, nz-1
        do i = 2, nx-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt(x(i)**2 + (y(j)+h)**2 + z(k)**2)
           u(i,j,k) = 1.0d0 + (u(i,j+1,k) - 1.0d0) * r_i / r_b
        enddo
      enddo
      
c------------------------------------------------------------
c +z face
c------------------------------------------------------------
      k = nz
      do j = 2, ny-1
        do i = 2, nx-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt(x(i)**2 + y(j)**2 + (z(k)-h)**2)
           u(i,j,k) = 1.0d0 + (u(i,j,k-1) - 1.0d0) * r_i / r_b
        enddo
      enddo
      
c------------------------------------------------------------
c -z face
c------------------------------------------------------------
      k = 1
      do j = 2, ny-1
        do i = 1, nx-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt(x(i)**2 + y(j)**2 + (z(k)+h)**2)
           u(i,j,k) = 1.0d0 + (u(i,j,k+1) - 1.0d0) * r_i / r_b
        enddo
      enddo


c=============================
c EDGES (12 of them)
c=============================
c------------------------------------------------------------
c +x +y edge
c------------------------------------------------------------
      i = nx      
      j = ny      
      do k = 2, nz-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)-h)**2 + (y(j)-h)**2 + z(k)**2)
           u(i,j,k) = 1.0d0 + (u(i-1,j-1,k) - 1.0d0) * r_i / r_b
      enddo

c------------------------------------------------------------
c +x -y edge
c------------------------------------------------------------
      i = nx      
      j = 1      
      do k = 2, nz-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)-h)**2 + (y(j)+h)**2 + z(k)**2)
           u(i,j,k) = 1.0d0 + (u(i-1,j+1,k) - 1.0d0) * r_i / r_b
      enddo
      
c------------------------------------------------------------
c -x +y edge: forward in x, backward in y, centered in z
c------------------------------------------------------------
      i = 1      
      j = ny      
      do k = 2, nz-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)+h)**2 + (y(j)-h)**2 + z(k)**2)
           u(i,j,k) = 1.0d0 + (u(i+1,j-1,k) - 1.0d0) * r_i / r_b
      enddo
      
c------------------------------------------------------------
c -x -y edge
c------------------------------------------------------------
      i = 1      
      j = 1      
      do k = 2, nz-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)+h)**2 + (y(j)+h)**2 + z(k)**2)
           u(i,j,k) = 1.0d0 + (u(i+1,j+1,k) - 1.0d0) * r_i / r_b
      enddo
      
      
c------------------------------------------------------------
c +x +z edge
c------------------------------------------------------------
      i = nx      
      k = nz
      do j = 2, ny-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)-h)**2 + y(j)**2 + (z(k)-h)**2)
           u(i,j,k) = 1.0d0 + (u(i-1,j,k-1) - 1.0d0) * r_i / r_b
      enddo

c------------------------------------------------------------
c +x -z edge
c------------------------------------------------------------
      i = nx      
      k = 1
      do j = 2, ny-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)-h)**2 + y(j)**2 + (z(k)+h)**2)
           u(i,j,k) = 1.0d0 + (u(i-1,j,k+1) - 1.0d0) * r_i / r_b
      enddo
      
c------------------------------------------------------------
c -x +z edge
c------------------------------------------------------------
      i = 1      
      k = nz
      do j = 2, ny-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)+h)**2 + y(j)**2 + (z(k)-h)**2)
           u(i,j,k) = 1.0d0 + (u(i+1,j,k-1) - 1.0d0) * r_i / r_b
      enddo
      
c------------------------------------------------------------
c -x -z edge
c------------------------------------------------------------
      i = 1      
      k = 1
      do j = 2, ny-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)+h)**2 + y(j)**2 + (z(k)+h)**2)
           u(i,j,k) = 1.0d0 + (u(i+1,j,k+1) - 1.0d0) * r_i / r_b
      enddo

      
c------------------------------------------------------------
c +y +z edge
c------------------------------------------------------------
      j = ny       
      k = nz
      do i = 2, nx-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt(x(i)**2 + (y(j)-h)**2 + (z(k)-h)**2)
           u(i,j,k) = 1.0d0 + (u(i,j-1,k-1) - 1.0d0) * r_i / r_b
      enddo

c------------------------------------------------------------
c +y -z edge
c------------------------------------------------------------
      j = ny       
      k = 1
      do i = 2, nx-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt(x(i)**2 + (y(j)-h)**2 + (z(k)+h)**2)
           u(i,j,k) = 1.0d0 + (u(i,j-1,k+1) - 1.0d0) * r_i / r_b
      enddo

c------------------------------------------------------------
c -y +z edge
c------------------------------------------------------------
      j = 1       
      k = nz
      do i = 2, nx-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt(x(i)**2 + (y(j)+h)**2 + (z(k)-h)**2)
           u(i,j,k) = 1.0d0 + (u(i,j+1,k-1) - 1.0d0) * r_i / r_b
      enddo

c------------------------------------------------------------
c -y -z edge
c------------------------------------------------------------
      j = 1       
      k = 1
      do i = 2, nx-1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt(x(i)**2 + (y(j)+h)**2 + (z(k)+h)**2)
           u(i,j,k) = 1.0d0 + (u(i,j+1,k+1) - 1.0d0) * r_i / r_b
      enddo


c=============================
c CORNERS (8 of them)
c=============================
c------------------------------------------------------------
c +x +y +z corner
c------------------------------------------------------------
      i = nx
      j = ny
      k = nz
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)-h)**2 + (y(j)-h)**2 + (z(k)-h)**2)
           u(i,j,k) = 1.0d0 + (u(i-1,j-1,k-1) - 1.0d0) * r_i / r_b
      
c------------------------------------------------------------
c +x +y -z corner
c------------------------------------------------------------
      i = nx
      j = ny
      k = 1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)-h)**2 + (y(j)-h)**2 + (z(k)+h)**2)
           u(i,j,k) = 1.0d0 + (u(i-1,j-1,k+1) - 1.0d0) * r_i / r_b
      
c------------------------------------------------------------
c +x -y +z corner
c------------------------------------------------------------
      i = nx
      j = 1
      k = nz
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)-h)**2 + (y(j)+h)**2 + (z(k)-h)**2)
           u(i,j,k) = 1.0d0 + (u(i-1,j+1,k-1) - 1.0d0) * r_i / r_b
      
c------------------------------------------------------------
c +x -y -z corner
c------------------------------------------------------------
      i = nx
      j = 1
      k = 1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)-h)**2 + (y(j)+h)**2 + (z(k)+h)**2)
           u(i,j,k) = 1.0d0 + (u(i-1,j+1,k+1) - 1.0d0) * r_i / r_b
      
c------------------------------------------------------------
c -x +y +z corner
c------------------------------------------------------------
      i = 1
      j = ny
      k = nz
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)+h)**2 + (y(j)-h)**2 + (z(k)-h)**2)
           u(i,j,k) = 1.0d0 + (u(i+1,j-1,k-1) - 1.0d0) * r_i / r_b
      
c------------------------------------------------------------
c -x +y -z corner
c------------------------------------------------------------
      i = 1
      j = ny
      k = 1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)+h)**2 + (y(j)-h)**2 + (z(k)+h)**2)
           u(i,j,k) = 1.0d0 + (u(i+1,j-1,k+1) - 1.0d0) * r_i / r_b
      
c------------------------------------------------------------
c -x -y +z corner
c------------------------------------------------------------
      i = 1
      j = 1
      k = nz
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)+h)**2 + (y(j)+h)**2 + (z(k)-h)**2)
           u(i,j,k) = 1.0d0 + (u(i+1,j+1,k-1) - 1.0d0) * r_i / r_b
      
c------------------------------------------------------------
c -x -y -z corner
c------------------------------------------------------------
      i = 1
      j = 1
      k = 1
           r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
           r_i = sqrt((x(i)+h)**2 + (y(j)+h)**2 + (z(k)+h)**2)
           u(i,j,k) = 1.0d0 + (u(i+1,j+1,k+1) - 1.0d0) * r_i / r_b
      
      
900   continue

      return
      end


c===========================================================
c rbc_dirichlet                             Scott Hawley
c
c  This doesn't really apply the Robin BC.  Rather we
c  apply a dirichlet condition of the values of the continuum
c  solution at the outer boundary...
c===========================================================
      subroutine rbc_dirichlet(u,nx,ny,nz,x,y,z,h,mask)
     
      implicit none

      include 'mgprob.inc'

      integer       nx,            ny,            nz
      real*8        u(nx,ny,nz)
      real*8        x(nx),         y(ny),         z(nz)
      real*8         h
      integer        mask(nx,ny,nz)


      integer        i,             j,            k
      real*8         r_i,           r_b

      call clcu2(u,x,z,y,nx,ny,nz,
     &           mask,.false.,.false.,.false.,.false.,.true.)

      return
      end



c===========================================================
c rbc_diffsonn                                   Scott Hawley
c
c  Apply Robin Boundary condition around outer boundary
c
c   The Robin condition is   
c
c                     du/dr = (1-u)/r
c
c    (where here and elsewhere d / dr is in fact a partial 
c    derivative), which isn't strictly linear, so we 
c    introduce a variable v = 1 - u, and solve
c
c                     dv/dr = -v/r
c
c      Thus we solve for this in Cartesian coordinates:

c             x dv/dx + y dv/dy + z dv/dz = -v
c
c                        or
c
c            v =  -x dv/dx - y dv/dy - z dv/dz 
c
c   Note that, although this equation is solved explicitly,
c   it assumes that these derivatives are KNOWN, which in general
c   they are NOT (because the values at the other grid points
c   on, e.g., a face of the cube, have not all been calculated
c   yet).  Thus some kind of iterative approach will be
c   necessary.
c   This routine will apply up to "npasses" passes/iterations
c   before exiting.
c
c   To save memory, instead of introducing storage
c   for the variable v, we simply use the array u itself 
c   inside this subroutine, i.e. we set u := 1 - u at the 
c   beginning and set u:= 1 - u at the end.
c
c   It would be faster to introduce storage for the 
c   coordinates x, y, and z, but we don't do so here.
c   We assume that the domain is [-0.5,-0.5,-0.5] to [0.5,0.5,0.5]
c   (can change this later...)
c===========================================================
      subroutine rbc_diffsonn(u,nx,ny,nz,x,y,z,h)
     
      implicit none

      integer       nx,            ny,            nz
      real*8        u(nx,ny,nz)
      real*8        x(nx),         y(ny),         z(nz)
      real*8        h

      integer     npasses
      parameter ( npasses = 1 )
      integer     pass

      integer        i,             j,            k


c------------------------------------------------------------
c   convert from u to "v" = 1 - u, using u's storage
c------------------------------------------------------------
      do k = 1, nz
         do j = 1, ny
            do i = 1, nx
               u(i,j,k) = 1.0d0 - u(i,j,k)
            enddo
         enddo
      enddo

c==================================================================
c     Make a total of npasses passes in apply(j)ing this BC.
c==================================================================

      do pass = 1, npasses

c==================================================================
c  Apply(j) boundary(j) condition on faces, edges, and corners of cube
c==================================================================

c=============================
c FACES (6 of them)
c=============================
c------------------------------------------------------------
c +x(i) face: backward in x(i), centered in y(j) & z(k)
c------------------------------------------------------------
      i = nx
      do k = 2, nz-1
        do j = 2, ny-1
           u(i,j,k) = - x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i-1,j,k) 
     &                            + u(i-2,j,k) ) / (2*h)
     &                - y(j) * (u(i,j+1,k) - u(i,j-1,k)) / (2*h)
     &                - z(k) * (u(i,j,k+1) - u(i,j,k-1)) / (2*h)
        enddo
      enddo

      
c------------------------------------------------------------
c -x(i) face: forward in x(i), centered in y(j) & z(k)
c------------------------------------------------------------
      i = 1
      do k = 2, nz-1
        do j = 2, ny-1
           u(i,j,k) = + x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i+1,j,k) 
     &                            + u(i+2,j,k) ) / (2*h)
     &                - y(j) * (u(i,j+1,k) - u(i,j-1,k)) / (2*h)
     &                - z(k) * (u(i,j,k+1) - u(i,j,k-1)) / (2*h)
        enddo
      enddo
      
c------------------------------------------------------------
c +y(j) face: backward in y(j), centered in x(i) & z(k)
c------------------------------------------------------------
      j = ny
      do k = 2, nz-1
        do i = 2, nx-1
           u(i,j,k) = - x(i) * (u(i+1,j,k) - u(i-1,j,k)) / (2*h)
     &                - y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j-1,k) 
     &                            + u(i,j-2,k) ) / (2*h)
     &                - z(k) * (u(i,j,k+1) - u(i,j,k-1)) / (2*h)
        enddo
      enddo
      
c------------------------------------------------------------
c -y(j) face: forward in y(j), centered in x(i) & z(k)
c------------------------------------------------------------
      j = 1
      do k = 2, nz-1
        do i = 2, nx-1
           u(i,j,k) = - x(i) * (u(i+1,j,k) - u(i-1,j,k)) / (2*h)
     &                + y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j+1,k) 
     &                            + u(i,j+2,k) ) / (2*h)
     &                - z(k) * (u(i,j,k+1) - u(i,j,k-1)) / (2*h)
        enddo
      enddo
      
c------------------------------------------------------------
c +z(k) face: backward in z(k), centered in x(i) & y(j)
c------------------------------------------------------------
      k = nz
      do j = 2, ny-1
        do i = 2, nx-1
           u(i,j,k) = - x(i) * (u(i+1,j,k) - u(i-1,j,k)) / (2*h)
     &                - y(j) * (u(i,j+1,k) - u(i,j-1,k)) / (2*h)
     &                - z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k-1) 
     &                            + u(i,j,k-2) ) / (2*h)
        enddo
      enddo
      
c------------------------------------------------------------
c -z(k) face: forward in z(k), centered in x(i) & y(j)
c------------------------------------------------------------
      k = 1
      do j = 2, ny-1
        do i = 1, nx-1
           u(i,j,k) = - x(i) * (u(i+1,j,k) - u(i-1,j,k)) / (2*h)
     &                - y(j) * (u(i,j+1,k) - u(i,j-1,k)) / (2*h)
     &                + z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k+1) 
     &                            + u(i,j,k+2) ) / (2*h)
        enddo
      enddo


c=============================
c EDGES (12 of them)
c=============================
c------------------------------------------------------------
c +x(i) +y(j) edge: backward in x(i) & y(j), centered in z(k)
c------------------------------------------------------------
      i = nx      
      j = ny      
      do k = 2, nz-1
           u(i,j,k) = - x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i-1,j,k) 
     &                            + u(i-2,j,k) ) / (2*h)
     &                - y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j-1,k) 
     &                            + u(i,j-2,k) ) / (2*h)
     &                - z(k) * (u(i,j,k+1) - u(i,j,k-1)) / (2*h)
      enddo

c------------------------------------------------------------
c +x(i) -y(j) edge: backward in x(i), forward in y(j), centered in z(k)
c------------------------------------------------------------
      i = nx      
      j = 1      
      do k = 2, nz-1
           u(i,j,k) = - x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i-1,j,k) 
     &                            + u(i-2,j,k) ) / (2*h)
     &                + y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j+1,k) 
     &                            + u(i,j+2,k) ) / (2*h)
     &                - z(k) * (u(i,j,k+1) - u(i,j,k-1)) / (2*h)
      enddo
      
c------------------------------------------------------------
c -x(i) +y(j) edge: forward in x(i), backward in y(j), centered in z(k)
c------------------------------------------------------------
      i = 1      
      j = ny      
      do k = 2, nz-1
           u(i,j,k) = + x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i+1,j,k) 
     &                            + u(i-2,j,k) ) / (2*h)
     &                - y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j-1,k) 
     &                            + u(i,j-2,k) ) / (2*h)
     &                - z(k) * (u(i,j,k+1) - u(i,j,k-1)) / (2*h)
      enddo
      
c------------------------------------------------------------
c -x(i) -y(j) edge: forward in x(i), forward in y(j), centered in z(k)
c------------------------------------------------------------
      i = 1      
      j = 1      
      do k = 2, nz-1
           u(i,j,k) = + x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i+1,j,k) 
     &                            + u(i-2,j,k) ) / (2*h)
     &                + y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j+1,k) 
     &                            + u(i,j+2,k) ) / (2*h)
     &                - z(k) * (u(i,j,k+1) - u(i,j,k-1)) / (2*h)
      enddo
      
      
c------------------------------------------------------------
c +x(i) +z(k) edge: backward in x(i), centered in y(j), backward in z(k)
c------------------------------------------------------------
      i = nx      
      k = nz
      do j = 2, ny-1
           u(i,j,k) = - x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i-1,j,k) 
     &                            + u(i-2,j,k) ) / (2*h)
     &                - y(j) * (u(i,j+1,k) - u(i,j-1,k)) / (2*h)
     &                - z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k-1) 
     &                            + u(i,j,k-2) ) / (2*h)
      enddo

c------------------------------------------------------------
c +x(i) -z(k) edge: backward in x(i), centered in y(j), forward in z(k)
c------------------------------------------------------------
      i = nx      
      k = 1
      do j = 2, ny-1
           u(i,j,k) = - x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i-1,j,k) 
     &                            + u(i-2,j,k) ) / (2*h)
     &                - y(j) * (u(i,j+1,k) - u(i,j-1,k)) / (2*h)
     &                + z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k+1) 
     &                            + u(i,j,k+2) ) / (2*h)
      enddo
      
c------------------------------------------------------------
c -x(i) +z(k) edge: forward in x(i), centered in y(j), backward in z(k)
c------------------------------------------------------------
      i = 1      
      k = nz
      do j = 2, ny-1
           u(i,j,k) = + x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i+1,j,k) 
     &                            + u(i+2,j,k) ) / (2*h)
     &                - y(j) * (u(i,j+1,k) - u(i,j-1,k)) / (2*h)
     &                - z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k-1) 
     &                            + u(i,j,k-2) ) / (2*h)
      enddo
      
c------------------------------------------------------------
c -x(i) -z(k) edge: forward in x(i), centered in y(j), forward in z(k)
c------------------------------------------------------------
      i = 1      
      k = 1
      do j = 2, ny-1
           u(i,j,k) = + x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i+1,j,k) 
     &                            + u(i+2,j,k) ) / (2*h)
     &                - y(j) * (u(i,j+1,k) - u(i,j-1,k)) / (2*h)
     &                + z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k+1) 
     &                            + u(i,j,k+2) ) / (2*h)
      enddo

      
c------------------------------------------------------------
c +y(j) +z(k) edge: centered in x(i), backward in y(j), backward in z(k)
c------------------------------------------------------------
      j = ny       
      k = nz
      do i = 2, nx-1
           u(i,j,k) = - x(i) * (u(i+1,j,k) - u(i-1,j,k)) / (2*h)
     &                - y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j-1,k) 
     &                            + u(i,j-2,k) ) / (2*h)
     &                - z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k-1) 
     &                            + u(i,j,k-2) ) / (2*h)
      enddo

c------------------------------------------------------------
c +y(j) -z(k) edge: centered in x(i), backward in y(j), forward in z(k)
c------------------------------------------------------------
      j = ny       
      k = 1
      do i = 2, nx-1
           u(i,j,k) = - x(i) * (u(i+1,j,k) - u(i-1,j,k)) / (2*h)
     &                - y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j-1,k) 
     &                            + u(i,j-2,k) ) / (2*h)
     &                + z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k+1) 
     &                            + u(i,j,k+2) ) / (2*h)
      enddo

c------------------------------------------------------------
c -y(j) +z(k) edge: centered in x(i), forward in y(j), backward in z(k)
c------------------------------------------------------------
      j = 1       
      k = nz
      do i = 2, nx-1
           u(i,j,k) = - x(i) * (u(i+1,j,k) - u(i-1,j,k)) / (2*h)
     &                + y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j+1,k) 
     &                            + u(i,j+2,k) ) / (2*h)
     &                - z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k-1) 
     &                            + u(i,j,k-2) ) / (2*h)
      enddo

c------------------------------------------------------------
c -y(j) -z(k) edge: centered in x(i), forward in y(j), forward in z(k)
c------------------------------------------------------------
      j = 1       
      k = 1
      do i = 2, nx-1
           u(i,j,k) = - x(i) * (u(i+1,j,k) - u(i-1,j,k)) / (2*h)
     &                + y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j+1,k) 
     &                            + u(i,j+2,k) ) / (2*h)
     &                + z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k+1) 
     &                            + u(i,j,k+2) ) / (2*h)
      enddo


c=============================
c CORNERS (8 of them)
c=============================
c------------------------------------------------------------
c +x(i) +y(j) +z(k) corner
c------------------------------------------------------------
      i = nx
      j = ny
      k = nz
           u(i,j,k) = - x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i-1,j,k) 
     &                            + u(i-2,j,k) ) / (2*h)
     &                - y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j-1,k) 
     &                            + u(i,j-2,k) ) / (2*h)
     &                - z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k-1) 
     &                            + u(i,j,k-2) ) / (2*h)
      
c------------------------------------------------------------
c +x(i) +y(j) -z(k) corner
c------------------------------------------------------------
      i = nx
      j = ny
      k = 1
           u(i,j,k) = - x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i-1,j,k) 
     &                            + u(i-2,j,k) ) / (2*h)
     &                - y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j-1,k) 
     &                            + u(i,j-2,k) ) / (2*h)
     &                + z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k+1) 
     &                            + u(i,j,k+-2) ) / (2*h)
      
c------------------------------------------------------------
c +x(i) -y(j) +z(k) corner
c------------------------------------------------------------
      i = nx
      j = 1
      k = nz
           u(i,j,k) = - x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i-1,j,k) 
     &                            + u(i-2,j,k) ) / (2*h)
     &                + y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j+1,k) 
     &                            + u(i,j+2,k) ) / (2*h)
     &                - z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k-1) 
     &                            + u(i,j,k-2) ) / (2*h)
      
c------------------------------------------------------------
c +x(i) -y(j) -z(k) corner
c------------------------------------------------------------
      i = nx
      j = 1
      k = 1
           u(i,j,k) = - x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i-1,j,k) 
     &                            + u(i-2,j,k) ) / (2*h)
     &                + y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j+1,k) 
     &                            + u(i,j+2,k) ) / (2*h)
     &                + z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k+1) 
     &                            + u(i,j,k+2) ) / (2*h)
      
c------------------------------------------------------------
c -x(i) +y(j) +z(k) corner
c------------------------------------------------------------
      i = 1
      j = ny
      k = nz
           u(i,j,k) = + x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i+1,j,k) 
     &                            + u(i+2,j,k) ) / (2*h)
     &                - y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j-1,k) 
     &                            + u(i,j-2,k) ) / (2*h)
     &                - z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k-1) 
     &                            + u(i,j,k-2) ) / (2*h)
      
c------------------------------------------------------------
c -x(i) +y(j) -z(k) corner
c------------------------------------------------------------
      i = 1
      j = ny
      k = 1
           u(i,j,k) = + x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i+1,j,k) 
     &                            + u(i+2,j,k) ) / (2*h)
     &                - y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j-1,k) 
     &                            + u(i,j-2,k) ) / (2*h)
     &                + z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k+1) 
     &                            + u(i,j,k+2) ) / (2*h)
      
c------------------------------------------------------------
c -x(i) -y(j) +z(k) corner
c------------------------------------------------------------
      i = 1
      j = 1
      k = nz
           u(i,j,k) = + x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i+1,j,k) 
     &                            + u(i+2,j,k) ) / (2*h)
     &                + y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j+1,k) 
     &                            + u(i,j+2,k) ) / (2*h)
     &                - z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k-1) 
     &                            + u(i,j,k-2) ) / (2*h)
      
c------------------------------------------------------------
c -x(i) -y(j) -z(k) corner
c------------------------------------------------------------
      i = 1
      j = 1
      k = 1
           u(i,j,k) = + x(i) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i+1,j,k) 
     &                            + u(i+2,j,k) ) / (2*h)
     &                + y(j) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j+1,k) 
     &                            + u(i,j+2,k) ) / (2*h)
     &                + z(k) * ( 3.0d0*u(i,j,k) - 4.0d0*u(i,j,k+1) 
     &                            + u(i,j,k+2) ) / (2*h)
      
      
c==============================================================
c     end of "passes" loop
c==============================================================
      end do

900   continue
c------------------------------------------------------------
c    convert back from "v" to u
c------------------------------------------------------------
      do k = 1, nz
         do j = 1, ny
            do i = 1, nx
               u(i,j,k) = 1.0d0 - u(i,j,k)
            enddo
         enddo
      enddo

      return
      end
      
      
      
      
      
      
c===========================================================
c rbc_project                                   Scott Hawley
c
c  Apply Robin Boundary condition around outer boundary
c
c   The Robin condition is   
c
c                     [r(u-1)],dr = 0
c
c    (where here and elsewhere d / dr is in fact a partial 
c    derivative), which isn't strictly linear, so we 
c    introduce a variable v = u - 1, and solve
c
c                     (rv),dr = 0
c
c===========================================================
      subroutine rbc_project(u,nx,ny,nz,x,y,z,h)
     
      implicit none

c    functions
      integer        mysgn
      real*8         linterp

      integer        nx,            ny,            nz
      real*8         u(nx,ny,nz)
      real*8         x(nx),         y(ny),        z(nz)
      real*8         h



c Local Variables

      integer        i,             j,            k
      real*8         dx,            dy,           dz

c   Variables for projection & interpolation
c     Direction towards the origin (+/-)
      integer        dirx,          diry,         dirz
c     Coordinates of projected point q
      real*8         xq,            yq,           zq
      real*8         rp,            rq
      real*8         r_b
      real*8         u_int1a,       u_int1b,      uq

      logical        ltrace
      parameter    ( ltrace = .false. )



      if ((nx .ne. ny) .or. (nx .ne. nz) .or. (ny .ne. nz)) then
         write(0,*) 'Error: Domain must be a cube.'
         return
      endif


c==================================================================
c  Apply boundary condition on faces, edges, and corners of cube
c==================================================================

c==================================================================
c FACES (6 of them)
c==================================================================
c------------------------------------------------------------
c   convert from u to "v" = 1 - u, using u's storage
c------------------------------------------------------------
      do k = 1, nz
         do j = 1, ny
            do i = 1, nx
               u(i,j,k) = u(i,j,k) - 1.0d0
            enddo
         enddo
      enddo


c------------------------------------------------------------
c +/- x faces
c------------------------------------------------------------
      do i = 1, nx, nx-1
        do k = 2, nz-1
          do j = 2, ny-1
          
            dirx = mysgn(-x(i))
            diry = mysgn(-y(j))
            dirz = mysgn(-z(k))

            xq = x(i+dirx)
            yq =  y(j) * xq / x(i) 
            zq =  z(k) * xq / x(i) 

            u_int1a = linterp( y(j),      u(i+dirx, j,     k), 
     &                         y(j+diry), u(i+dirx, j+diry,k), yq )
            u_int1b = linterp( y(j),      u(i+dirx, j,     k+dirz), 
     &                         y(j+diry), u(i+dirx, j+diry,k+dirz), yq )
            uq      = linterp( z(k),      u_int1a,
     &                         z(k+dirz), u_int1b, zq )

            rq = sqrt( xq**2 + yq**2 + zq**2 )
            r_b = sqrt( x(i)**2 + y(j)**2 + z(k)**2 )
c            u(i,j,k) = 1.062d0 * uq * rq / r_b
            u(i,j,k) = uq * rq / r_b

            if (ltrace) then
              write(0,*) ' '
              write(0,*) 'i, j, k           = ', i,j,k
              write(0,*) 'dirx, diry, dirz  = ', dirx,diry,dirz
              write(0,*) 'x(i),y(j),z(k),r_b= ',x(i), y(j), z(k),r_b
              write(0,*) 'xq, yq, zq, rq    = ', xq, yq, zq, rq
              write(0,*)'u(i+dirx, j, k),    u(i+dirx, j+diry,k)    =',
     &                    u(i+dirx, j, k),      u(i+dirx, j+diry,k)
              write(0,*)'u(i+dirx,j,k+dirz), u(i+dirx,j+diry,k+dirz)=',
     &                    u(i+dirx, j, k+dirz), u(i+dirx, j+diry,k+dirz)
              write(0,*) 'u_int1a, u_int1b  = ',u_int1a, u_int1b
              write(0,*) 'uq, u_proj        = ',uq, u(i,j,k)
              write(0,*) 'uqperp, u_perpprj =', u(i+dirx,j,k), 
     &         u(i+dirx,j,k) * sqrt(x(i+dirx)**2 +y(j)**2 +z(k)**2)/ r_b
              write(0,*) 'uqperp+1 = ', u(i+dirx,j+diry,k+dirz)
              write(0,*) 'uexact(q), uexact(b) = ',
     &                    1.0d0 - (1.0d0 - 2.0d0 / rq),
     &                    1.0d0 - (1.0d0 - 2.0d0 / r_b)
            endif

            
            u(i,j,k) = u(i+dirx,j,k) * 
     &              sqrt(x(i+dirx)**2 +y(j)**2 +z(k)**2) / r_b
          enddo
        enddo
      enddo
      


c-----------------------------------------------------------
c +/- y faces
c------------------------------------------------------------
      do j = 1, ny, ny-1
        do k = 2, nz-1
          do i = 2, nx-1
            dirx = mysgn(-x(i))
            diry = mysgn(-y(j))
            dirz = mysgn(-z(k))

            yq = y(j+diry)
            xq = yq * x(i) / y(j) 
            zq = yq * z(k) / y(j) 

            u_int1a = linterp( x(i),      u(i,      j+diry, k), 
     &                         x(i+dirx), u(i+dirx, j+diry, k), xq )
            u_int1b = linterp( x(i),      u(i,      j+diry, k+dirz), 
     &                         x(i+dirx), u(i+dirx, j+diry, k+dirz), xq)
            uq      = linterp( z(k),      u_int1a,
     &                         z(k+dirz), u_int1b, zq )

            rq = sqrt( xq**2 + yq**2 + zq**2 )
            r_b = sqrt( x(i)**2 + y(j)**2 + z(k)**2 )
            u(i,j,k) = uq * rq / r_b

            u(i,j,k) = u(i,j+diry,k) * 
     &              sqrt(x(i)**2 +y(j+diry)**2 +z(k)**2) / r_b
          enddo
        enddo
      enddo


c------------------------------------------------------------
c +/- z faces
c------------------------------------------------------------
      do k = 1, nz, nz-1
        do j = 2, ny-1
          do i = 2, nx-1
            dirx = mysgn(-x(i))
            diry = mysgn(-y(j))
            dirz = mysgn(-z(k))

            zq = z(k+dirz)
            xq = zq * x(i) / z(k) 
            yq = zq * y(j) / z(k) 

            u_int1a = linterp( x(i),      u(i,      j,      k+dirz), 
     &                         x(i+dirx), u(i+dirx, j,      k+dirz), xq)
            u_int1b = linterp( x(i),      u(i,      j+diry, k+dirz), 
     &                         x(i+dirx), u(i+dirx, j+diry, k+dirz), xq)
            uq      = linterp( y(j),      u_int1a,
     &                         y(j+diry), u_int1b, yq )

            rq = sqrt( xq**2 + yq**2 + zq**2 )
            r_b = sqrt( x(i)**2 + y(j)**2 + z(k)**2 )
            u(i,j,k) = uq * rq / r_b

            u(i,j,k) = u(i,j,k+dirz) * 
     &              sqrt(x(i)**2 +y(j)**2 +z(k+dirz)**2) / r_b
          enddo
        enddo
      enddo

      
c=============================
c EDGES (12 of them)
c=============================
c------------------------------------------------------------
c +/- x,  +/- y edges
c------------------------------------------------------------
      do i = 1, nx, nx-1
        do j = 1, ny, ny-1
           do k = 2, nz-1
             dirx = mysgn(-x(i))
             diry = mysgn(-y(j))
             dirz = mysgn(-z(k))

             xq = x(i+dirx)
             yq = y(j+diry) 
             zq = yq * z(k) / y(j) 

             uq     = linterp( z(k),      u(i+dirx, j+diry, k),
     &                         z(k+dirz), u(i+dirx, j+diry, k+dirz), zq)

             rq = sqrt( xq**2 + yq**2 + zq**2 )
             r_b = sqrt( x(i)**2 + y(j)**2 + z(k)**2 )
             u(i,j,k) = uq * rq / r_b
           enddo
         enddo
      enddo

c------------------------------------------------------------
c +/- x +/- z edges
c------------------------------------------------------------
      do i = 1, nx, nx-1
        do k = 1, nz, nz-1
          do j = 2, ny-1
            dirx = mysgn(-x(i))
            diry = mysgn(-y(j))
            dirz = mysgn(-z(k))

             xq = x(i+dirx)
             yq = xq * y(j) / x(i) 
             zq = z(k+dirz) 

             uq     = linterp( y(j),      u(i+dirx, j,      k+dirz),
     &                         y(j+diry), u(i+dirx, j+diry, k+dirz), yq)

             rq = sqrt( xq**2 + yq**2 + zq**2 )
             r_b = sqrt( x(i)**2 + y(j)**2 + z(k)**2 )
             u(i,j,k) = uq * rq / r_b
          enddo
        enddo
      enddo

c------------------------------------------------------------
c +/- y, +/- z edges
c------------------------------------------------------------
      do j = 1, ny, ny-1
        do k = 1, nz, nz-1
          do i = 2, nx-1
            dirx = mysgn(-x(i))
            diry = mysgn(-y(j))
            dirz = mysgn(-z(k))

             yq = y(j+diry)
             zq = z(k+dirz) 
             xq = yq * x(i) / y(j) 

             uq     = linterp( x(i),      u(i,      j+diry, k+dirz),
     &                         x(i+dirx), u(i+dirx, j+diry, k+dirz), xq)

             rq = sqrt( xq**2 + yq**2 + zq**2 )
             r_b = sqrt( x(i)**2 + y(j)**2 + z(k)**2 )
             u(i,j,k) = uq * rq / r_b
          enddo
        enddo
      enddo

c=============================
c CORNERS (8 of them)
c=============================
      do k = 1, nz, nz-1
        do j = 1, ny, ny-1
          do i = 1, nx, nx-1
            dirx = mysgn(-x(i))
            diry = mysgn(-y(j))
            dirz = mysgn(-z(k))

            uq     = u(i+dirx,j+diry,k+dirz)

            r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
            rq = sqrt(x(i+dirx)**2 + y(j+diry)**2 + z(k+dirz)**2)
            u(i,j,k) = uq * rq / r_b
          enddo
        enddo
      enddo
      
c------------------------------------------------------------
c    convert back from "v" to u
c------------------------------------------------------------
      do k = 1, nz
         do j = 1, ny
            do i = 1, nx
               u(i,j,k) = u(i,j,k) + 1.0d0 
            enddo
         enddo
      enddo

      
900   continue

      return
      end


c===========================================================
c   rbc_2ocopy                                   Scott Hawley
c
c   The Robin condition is   
c
c                     partial u / partial r = (1-u)/r
c
c                           or
c
c                   [ r (u-1) ],r = 0
c
c  We use second-order forwards (or backwards) derivatives
c  in the direction normal to the faces of the cube...
c  letting v = u - 1, we have, at a boundary point (i,j,k)
c
c     3 r_i v_i  - 4 r_(i-1) v_(i-1) + r_(i-2) v_(i-2) = 0
c
c                          or
c
c v_i  =  ( 4 r_(i-1) v_(i-1) - r_(i-2) v_(i-2) ) / (3 r_i)
c 
c===========================================================
      subroutine rbc_2ocopy(u,nx,ny,nz,x,y,z,h
     &                  )
     
      implicit none
 
      integer       mysgn

      integer       nx,            ny,            nz
      real*8        u(nx,ny,nz)
      real*8        x(nx),        y(ny),       z(nz)
      real*8         h

      integer        i,             j,            k
      integer        dirx,          diry,         dirz
      real*8         r_b,           r_in,         r_in2

c=============================
c FACES (6 of them)
c=============================
c------------------------------------------------------------
c    convert from u to "v" 
c------------------------------------------------------------
      do k = 1, nz
         do j = 1, ny
            do i = 1, nx
               u(i,j,k) = 1.0d0 - u(i,j,k)
            enddo
         enddo
      enddo

c------------------------------------------------------------
c (interior of) +/- x faces
c------------------------------------------------------------
      do i = 1, nx, nx-1
        do k = 2, nz-1  
          do j = 2, ny-1
            dirx = -mysgn(x(i))
            diry = -mysgn(y(j))
            dirz = -mysgn(z(k))
            
            r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
            r_in = sqrt((x(i+dirx))**2 + y(j)**2 + z(k)**2)
            r_in2 = sqrt((x(i+2*dirx))**2 + y(j)**2 + z(k)**2)
            
            u(i,j,k) = (4* r_in * u(i+dirx,j,k) - 
     &                     r_in2 * u(i+2*dirx,j,k)) / (3 * r_b)
          enddo
        enddo
      enddo
      
     
c-----------------------------------------------------------
c +/- y faces
c------------------------------------------------------------
      do j = 1, ny, ny-1
        do k = 2, nz-1
          do i = 2, nx-1
            dirx = -mysgn(x(i))
            diry = -mysgn(y(j))
            dirz = -mysgn(z(k))

            r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
            r_in = sqrt(x(i)**2 + y(j+diry)**2 + z(k)**2)
            r_in2 = sqrt(x(i)**2 + y(j+2*diry)**2 + z(k)**2)

            u(i,j,k) = (4* r_in * u(i,j+diry,k) - 
     &                     r_in2 * u(i,j+2*diry,k)) / (3 * r_b)
          enddo
        enddo
      enddo

      
c------------------------------------------------------------
c +/- z faces
c------------------------------------------------------------
      do k = 1, nz, nz-1
        do j = 2, ny-1
          do i = 2, nx-1
            dirx = -mysgn(x(i))
            diry = -mysgn(y(j))
            dirz = -mysgn(z(k))
            r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
            r_in = sqrt(x(i)**2 + y(j)**2 + z(k+dirz)**2)
            r_in2 = sqrt(x(i)**2 + y(j)**2 + z(k+2*dirz)**2)

            u(i,j,k) = (4* r_in * u(i,j,k+dirz) - 
     &                     r_in2 * u(i,j,k+2*dirz)) / (3 * r_b)
          enddo
        enddo
      enddo
      


c=============================
c EDGES (12 of them)
c=============================
c------------------------------------------------------------
c +x +y edge
c------------------------------------------------------------
      do i = 1, nx, nx-1
        do j = 1, ny, ny-1
          do k = 2, nz-1
            dirx = -mysgn(x(i))
            diry = -mysgn(y(j))
            dirz = -mysgn(z(k))
            
            r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
            r_in = sqrt((x(i+dirx))**2 + y(j+diry)**2 + z(k)**2)
            r_in2 = sqrt((x(i+2*dirx))**2 + y(j+2*diry)**2 + z(k)**2)
            
            u(i,j,k) = (4* r_in * u(i+dirx,j+diry,k) - 
     &                     r_in2 * u(i+2*dirx,j+2*diry,k)) / (3 * r_b)
          enddo
        enddo
      enddo


c------------------------------------------------------------
c +/- x, +/- z edges
c------------------------------------------------------------
      do i = 1, nx, nx-1
        do k = 1, nz, nz-1
          do j = 2, ny-1
            dirx = -mysgn(x(i))
            diry = -mysgn(y(j))
            dirz = -mysgn(z(k))
            
            r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
            r_in = sqrt(x(i+dirx)**2 + y(j)**2 + z(k+dirz)**2)
            r_in2 = sqrt(x(i+2*dirx)**2 + y(j)**2 + z(k+2*dirz)**2)
            
            u(i,j,k) = (4* r_in * u(i+dirx,j,k+dirz) - 
     &                     r_in2 * u(i+2*dirx,j,k+2*dirz)) / (3 * r_b)
          enddo
        enddo
      enddo

      
c------------------------------------------------------------
c +/- y, +/- z edges
c------------------------------------------------------------
      do k = 1, nz, nz-1
        do j = 1, ny, ny-1
          do i = 2, nx-1
            dirx = -mysgn(x(i))
            diry = -mysgn(y(j))
            dirz = -mysgn(z(k))
            
            r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
            r_in = sqrt(x(i)**2 + y(j+diry)**2 + z(k+dirz)**2)
            r_in2 = sqrt(x(i)**2 + y(j+2*diry)**2 + z(k+2*dirz)**2)
            
            u(i,j,k) = (4* r_in * u(i,j+diry,k+dirz) - 
     &                     r_in2 * u(i,j+2*diry,k+2*dirz)) / (3 * r_b)
          enddo
        enddo
      enddo

c=============================
c CORNERS (8 of them)
c=============================
      do k = 1, nz, nz-1
        do j = 1, ny, ny-1
          do i = 1, nx, nx-1
            dirx = -mysgn(x(i))
            diry = -mysgn(y(j))
            dirz = -mysgn(z(k))

            r_b = sqrt(x(i)**2 + y(j)**2 + z(k)**2)
            r_in = sqrt(x(i+dirx)**2 + y(j+diry)**2 + z(k+dirz)**2)
            r_in2 = sqrt(x(i+2*dirx)**2 + y(j+2*diry)**2 
     &                 + z(k+2*dirz)**2)

            u(i,j,k) = (4* r_in * u(i+dirx,j+diry,k+dirz) - 
     &          r_in2 * u(i+2*dirx,j+2*diry,k+2*dirz)) / (3 * r_b)
          enddo
        enddo
      enddo
      
c------------------------------------------------------------
c    convert back from "v" to u
c------------------------------------------------------------
      do k = 1, nz
         do j = 1, ny
            do i = 1, nx
               u(i,j,k) = 1.0d0 - u(i,j,k)
            enddo
         enddo
      enddo
      
900   continue

      return
      end




c===========================================================
c  linterp    Simple 1D linear interpolation
c===========================================================
      real*8 function linterp ( x1, y1, x2, y2, x )
  
      implicit none

      real*8   x1, y1, x2, y2, x
      real*8   tmp

      if (x2 .ne. x1)  then
         tmp = y1 + (x - x1) * (y2 - y1) / (x2 - x1) 
         linterp = tmp
      else
         write(0,*) 'linterp: Assertion x2 .ne. x1 failed.'
         stop
      endif 

c      write(0,*) 'linterp: x1, y1, x2, y2, x, tmp = ',x1,y1,x2,y2,x,tmp

      return
      end

c===========================================================
c  mysgn      Simple sign function, but gives +1 if val = 0
c===========================================================
      integer  function mysgn ( val )
  
      implicit none

      real*8 val

      if (val .ne. 0.0d0) then
         mysgn = val / abs(val)
      else
         mysgn = 1
      endif

      return
      end
